### Jan 2011
###
### Code for approximation functions

negative.log.posterior <- function(theta, data)
{
# theta is a vector that contains the d parameters
  - generalized.log.posterior(theta, data)
}

# The maximum posterior estimator can be found using the following:

#nlmdetails <- nlm(negative.log.posterior, theta0, data=data, hessian = TRUE)
#thetahat <- nlmdetails$estimate
#J <- nlmdetails$hessian


# define local functions, second.derivative and j.theta, these will be used
# to calculate j (observed information), sigma (variance-covariance matrix),
# nu(theta), posterior expectations and predictive densities using T&k formula.


second.derivative <- function(f, theta, ...)
{
# theta is a vector that contains the d parameters
# calculation of 2nd derivative matrix of the function at the vector theta
# f is the score function

    epsilon <- 0.001
    del <- epsilon/2
    d <- length(theta)
    m <- matrix(0, ncol = d, nrow = d)  # cat(m, "\n")
    for(i in 1:d) {
        for(j in 1:d) {
            diff <- seq(0, 0, len = d)
            diff[j] <- del
            m[i, j] <- (f(theta + diff, ...)[i] - f(theta - diff, ...)[i])/
                epsilon #   cat(m[i, j], "\n")
        }
    }
    m

}# end of local function second.derivative.


j.theta <- function(f, theta, ...)
    {
# calculation the negative 2nd derivative matrix
# of the score function at the vector theta
# f is the score function

 - second.derivative(f, theta, ...)

}# end of local function j.theta.


# j (observed information) at maximum thetahat can be obtained
#J <- j.theta(generalized.score.posterior,thetahat,data)


# Define local functions, local.log.posterior, negative.profile.posterior. These
# functions return the conditional maximizer of generalized log-posterior function.

local.log.posterior <- function(lasttheta, subtheta, thetahat, data)
{
# subtheta is the vector of first i components.
# lasttheta is the vector of remaining (d-i) components over
# which the maximization takes place.

    theta <- c(subtheta, lasttheta)
    - generalized.log.posterior(theta, data)  # (generalized) log-posterior function

}# end of local function local.log.posterior.


negative.profile.posterior <- function(subtheta, thetahat, data)
{
# subtheta is the vector of first i components.

    i <- length(subtheta)
    j <- i + 1
    lasttheta0 <- thetahat[j:d]
    nlm(local.log.posterior, lasttheta0, subtheta = subtheta, thetahat = thetahat, data=data)

}# end of local function negative.profile.posterior.


# Define local function main.vector.theta, this function returns the
# p-vector c(thetahat_(i-1), theta, thetahat^(i+1)), where thetahat^(i+1)
# is the conditional maximizer of generalized log-likelihood function.

main.vector.theta <- function(thetai, i, thetahat, data)
{
# thetai is the i-th component.

    subtheta0 <- thetahat[1:(i - 1)]
    subtheta <- if(i > 1) c(subtheta0, thetai) else thetai
    condmax <- if(i < d) negative.profile.posterior(subtheta, thetahat, data)$est else subtheta
# condmax is the conditional maximizer of the log-posterior function.
    if(i < d)
        c(subtheta, condmax)
    else subtheta

}# end of local function main.vector.theta.


signed.root.inverse <- function(r, i, thetahat, data, J)
{
# estimating the parameter using z-formula.
# r is an arbitrary constant.
# thetahat is the MLE.
# J is the observed information

    zz <- r
    inverse <- solve(J[i:d, i:d])
    s <- 1/inverse[1, 1]
    h <- thetahat[i]
    thetai <- h + (zz/sqrt(s))
    main.vector.theta(thetai, i, thetahat, data)
}


thetaplusminus <- function(r, thetahat, data, J)
{
# r is an arbitrary constant.

    thetaplusminus <- matrix(0, ncol = d, nrow = d)
    for(i in 1:d) {
        thetaplusminus[i,  ] <- signed.root.inverse(r, i, thetahat, data, J)
    }
    thetaplusminus

}# end of function thetaplusminus.

# Running the above function we get
#thetaminus <- thetaplusminus(-1, thetahat, data, J)# matrix of thetaminus.
#thetaplus <- thetaplusminus(1, thetahat, data, J)# matrix of thetaplus.


# Define local function nu which will be used to compute tau,
# alphaminus and alphaplus

nu <- function(theta, i, data)
{
# calculating nu at a given vector theta.
# Define lambda, the (generalized) prior function.

    d <- length(theta)
    j <- (i + 1)
    sub.mat <- if(j < (d + 1)) j.theta(generalized.score.posterior,theta,data)[j:d, j:d
            ] else 1
    (det(sub.mat))^(-1/2)

}# end of local function nu.

# Define function alphaplusminus that computes alphaminus and alphaplus:

alphaplusminus <- function(thetaminus, thetaplus, data)
{

    alphaminus <- rep(0, len = d)
    alphaplus <- rep(0, len = d)
    for(i in 1:d) {
        numinus <- nu(thetaminus[i,  ], i, data)
        nuplus <- nu(thetaplus[i,  ], i, data)
        scoreminus <- generalized.score.posterior(thetaminus[i,  ], data)[i]
        scoreplus <- generalized.score.posterior(thetaplus[i,  ], data)[i]
        ratiominus <- numinus/scoreminus
        ratioplus <-  - nuplus/scoreplus
        tau <- ratiominus + ratioplus
        alphaminus[i] <- ratiominus/tau
        alphaplus[i] <- ratioplus/tau
    }
    list(alphaminus = alphaminus, alphaplus = alphaplus)
}# end of function alphaplusminus.

#local <- alphaplusminus(thetaminus, thetaplus, data)
#alphaminus<- local$alphaminus  # vector of alphaminus
#alphaplus <- local$alphaplus    # vector of alphaplus


# define function SKexpectation, for computing posterior
# expectation using Sweeting and kharroubi's formula

SKexpectation <- function(v, lambda, thetahat, thetaminus, thetaplus, alphaminus, alphaplus)
{
# v is a real-valued function of theta.

    vminus <- rep(0, len = d)
    vplus <- rep(0, len = d)
    vhat <- v(thetahat, lambda)
    for(i in 1:d) {
        vminus[i] <- v(thetaminus[i,  ], lambda)
        vplus[i] <- v(thetaplus[i,  ], lambda)
    }
    expect <- vhat + sum((alphaminus * vminus) + (alphaplus * vplus) - vhat)
    list(firstorder = vhat, secondorder = expect)

}# end of function SKexpectation.
